package request

import (
	"context"
	"crypto/tls"
	"fmt"
	"log"
	"query-service/internal/entity"

	"encoding/json"
	"io/ioutil"

	driver "github.com/arangodb/go-driver"
	"github.com/arangodb/go-driver/http"
)

//attr interface{}

//map[1 , 2 , 3 map [ .. ]]

/*
SendAQLQuery send AQL string query to database and returns a JSON object in a general format

Parameters: AQLQuery is a string containing the query that will be send to the database

Return: a map with two entries: "nodes" with a list of vertices/nodes and "edges" with a list of edges
that will be returned to the frontend
*/
func (s *Service) SendAQLQuery(query string, username string, password string, hostname string, port int, database string) (*[]byte, error) {

	//(*map[string][]entity.Document, error) = old return
	var queryResult = make(map[string][]entity.Document)
	conn, err := http.NewConnection(http.ConnectionConfig{
		Endpoints: []string{fmt.Sprintf("%s:%d", hostname, port)},
		TLSConfig: &tls.Config{InsecureSkipVerify: true},
	})
	if err != nil {
		log.Println("could not connect to database") // Handle error
		return nil, err
	}
	c, err := driver.NewClient(driver.ClientConfig{
		Connection:     conn,
		Authentication: driver.BasicAuthentication(username, password),
	})
	if err != nil {
		log.Println("Could not log in to the ArangoDB") // Handle error
		return nil, err
	}

	ctx := context.Background()
	db, err := c.Database(ctx, database)
	if err != nil {
		// handle error
		log.Println(err)
		return nil, err
	}

	cursor, err := db.Query(ctx, query, nil)
	if err != nil {
		log.Println("Invalid query") // handle error
		return nil, err
	}
	defer cursor.Close()

	//Loop through the resulting documents
	listContainer := entity.ListContainer{}
	num := -1.0
	isNum := false

	for {
		//var doc map[string][]interface{}
		var doc interface{}
		_, err := cursor.ReadDocument(ctx, &doc)
		if driver.IsNoMoreDocuments(err) {
			break
		} else if err != nil {
			// handle other errors
			fmt.Println(err)
			return nil, err
		}

		// Switch on the type of return, to filter out the case of a single number
		switch doc.(type) {
		case float64:
			num = doc.(float64)
			isNum = true
			break
		case map[string]interface{}:
			pdoc := doc.(map[string]interface{})
			parseResult(pdoc, &listContainer)
			break
		default:
			fmt.Println("Incompatible result type")
			break
		}
	}

	if !isNum {
		// Return nodes and edges
		queryResult["nodes"] = listContainer.NodeList
		queryResult["edges"] = listContainer.EdgeList

		jsonQ, err := json.Marshal(queryResult)
		return &jsonQ, err
	}
	// Return just a number
	numQ, err := json.Marshal(num)
	return &numQ, err

}

/* parseResult takes the result of the query and translates this to two lists: a nodelist and an edgelist, stored in a listcontainer

Parameters: doc is the document with the nodes and vertices that came back from the database,
listContainer is a struct containing the nodelist and edgelist that will be returned to the frontend

Return: Nothing because the result is stored in the listContainer
*/
func parseResult(doc map[string]interface{}, listContainer *entity.ListContainer) {

	//doc, ok := incomingDoc.(map[string][]interface{})
	vertices := doc["vertices"].([]interface{})
	edges := doc["edges"].([]interface{})

	for _, vertex := range vertices {
		vertexDoc := vertex.(map[string]interface{})

		(*listContainer).NodeList = append((*listContainer).NodeList, parseNode(vertexDoc))
	}

	for _, edge := range edges {
		edgeDoc := edge.(map[string]interface{})

		(*listContainer).EdgeList = append((*listContainer).EdgeList, parseEdge(edgeDoc))
	}
}

/* parseEdge parses the data of an edge to an output-friendly format
Parameters: d is a single entry of an edge

Return: a document with almost the same structure as before, but the attributes are grouped
*/
func parseEdge(d map[string]interface{}) entity.Document {
	doc := d //.(map[string]interface{})

	data := make(entity.Document)
	data["_id"] = doc["_id"]
	delete(doc, "_id")
	data["_key"] = doc["_key"]
	delete(doc, "_key")
	data["_rev"] = doc["_rev"]
	delete(doc, "_rev")
	data["_from"] = doc["_from"]
	delete(doc, "_from")
	data["_to"] = doc["_to"]
	delete(doc, "_to")
	data["attributes"] = doc

	//delete(doc, "_key")
	//data.rev = fmt.Sprintf("%v", doc["_rev"])

	// delete(doc, "_rev")
	// data.attr = doc
	return data
}

// writeJSON writes a json file for testing purposes
func writeJSON(queryResult map[string][]entity.Document) {
	file, _ := json.MarshalIndent(queryResult, "", " ")

	_ = ioutil.WriteFile("result.json", file, 0644)
}

/* parseEdge parses the data of a node to an output-friendly format
Parameters: d is a single entry of an node

Return: a document with almost the same structure as before, but the attributes are grouped
*/
func parseNode(d map[string]interface{}) entity.Document {
	doc := d //.(map[string]interface{})

	data := make(entity.Document)
	data["_id"] = doc["_id"]
	delete(doc, "_id")
	data["_key"] = doc["_key"]
	delete(doc, "_key")
	data["_rev"] = doc["_rev"]
	delete(doc, "_rev")
	data["attributes"] = doc

	//delete(doc, "_key")
	//data.rev = fmt.Sprintf("%v", doc["_rev"])

	// delete(doc, "_rev")
	// data.attr = doc
	return data
}
